import base64
import errno
import json
import socket
import threading
import traceback

from security.keystore import UserKeyStore

class RPCServer(object):
    handlers = {}

    @staticmethod
    def process_request(signed_request_encoded, keystore):
        try:
            signed_request = json.loads(base64.b64decode(signed_request_encoded))
            request = signed_request['request']
            signature = signed_request['signature']
            username = signed_request['username']
            signature_verified = keystore.rsa_verify_signature(request, signature, username)

            procedure_call = json.loads(base64.b64decode(request))
            procname = procedure_call['procname']
            args = procedure_call.get('args', [])
            kwargs = procedure_call.get('kwargs', {})
            if signature_verified:
                kwargs['rpc_signatory'] = username
            else:
                kwargs['rpc_signatory'] = None
            handler = RPCServer.handlers[procname]
            retval = handler(*args, **kwargs)
        except Exception as e:
            print traceback.format_exc()
            return RPCServer.return_error(e)
        return RPCServer.return_success(retval)
            
    @staticmethod
    def return_error(e):
        if hasattr(e, 'errno'):
            err = e.errno
        else:
            err = errno.ENOSYS
        return base64.b64encode(json.dumps({'status': 'error',
                                            'exception': {'type': str(type(e)),
                                                          'errno': err,
                                                          'val': str(e)}}))
    @staticmethod
    def return_success(retval):
        return base64.b64encode(json.dumps({'status': 'success',
                                            'retval': retval}))

    @staticmethod
    def register_handler(procname, handler):
        RPCServer.handlers[procname] = handler

    @staticmethod
    def start(host="0.0.0.0", port=9000, keystore=None):
        """
        Sits and listens to connections on a given host and port. Each
        incoming connection is handled by a separate thread.
        """
        def handle_thread(conn):
            rfile = conn.makefile('r')
            wfile = conn.makefile('w')
            request = ""
            while len(request) == 0:
                request = rfile.readline().strip()
            response = RPCServer.process_request(request, keystore)
            wfile.write("%s\n" % (response,))
            wfile.flush()
            r = rfile.read()
            if r == "":
                # socket is closed
                return
            else:
                print "Something really weird happened..."
                print r
                return

        def thread_factory(c):
            def thread_handler():
                return handle_thread(c)
            return thread_handler

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((host, port))
        sock.listen(5)
        while True:
            conn, addr = sock.accept()
            client_thread = threading.Thread(target=thread_factory(conn))
            client_thread.start()
